@@ -1,5 +1,5 @@
 //
-// Copyright 2021 Staysail Systems, Inc. <info@staysail.tech>
+// Copyright 2024 Staysail Systems, Inc. <info@staysail.tech>
 // Copyright 2018 Capitar IT Group BV <info@capitar.com>
 // Copyright 2018 Devolutions <info@devolutions.net>
 //
@@ -15,30 +15,13 @@
 #include <stdio.h>
 #include <string.h>
 
-// Functionality related to dialers.
+// Functionality related to dialing.
 static void dialer_connect_start(nni_dialer *);
 static void dialer_connect_cb(void *);
 static void dialer_timer_cb(void *);
 
-static nni_id_map dialers;
-static nni_mtx    dialers_lk;
-
-int
-nni_dialer_sys_init(void)
-{
-	nni_id_map_init(&dialers, 1, 0x7fffffff, false);
-	nni_mtx_init(&dialers_lk);
-
-	return (0);
-}
-
-void
-nni_dialer_sys_fini(void)
-{
-	nni_reap_drain();
-	nni_mtx_fini(&dialers_lk);
-	nni_id_map_fini(&dialers);
-}
+static nni_id_map dialers    = NNI_ID_MAP_INITIALIZER(1, 0x7fffffff, 0);
+static nni_mtx    dialers_lk = NNI_MTX_INITIALIZER;
 
 uint32_t
 nni_dialer_id(nni_dialer *d)
@@ -63,7 +46,7 @@ nni_dialer_destroy(nni_dialer *d)
 	NNI_FREE_STRUCT(d);
 }
 
-#if NNG_ENABLE_STATS
+#ifdef NNG_ENABLE_STATS
 static void
 dialer_stat_init(nni_dialer *d, nni_stat_item *item, const nni_stat_info *info)
 {
@@ -215,6 +198,9 @@ nni_dialer_bump_error(nni_dialer *d, int err)
 	case NNG_ENOMEM:
 		nni_stat_inc(&d->st_oom, 1);
 		break;
+	case NNG_ECLOSED:
+		// do nothing.
+		break;
 	default:
 		nni_stat_inc(&d->st_other, 1);
 		break;
@@ -225,15 +211,20 @@ nni_dialer_bump_error(nni_dialer *d, int err)
 #endif
 }
 
+// nni_dialer_create creates a dialer on the socket.
+// The caller should have a hold on the socket, and on success
+// the dialer inherits the callers hold.  (If the caller wants
+// an additional hold, it should get an extra hold before calling this
+// function.)
 int
-nni_dialer_create(nni_dialer **dp, nni_sock *s, const char *urlstr)
+nni_dialer_create(nni_dialer **dp, nni_sock *s, const char *url_str)
 {
-	nni_sp_tran *  tran;
-	nni_dialer *d;
-	int         rv;
-	nni_url *   url;
+	nni_sp_tran *tran;
+	nni_dialer  *d;
+	int          rv;
+	nni_url     *url;
 
-	if ((rv = nni_url_parse(&url, urlstr)) != 0) {
+	if ((rv = nni_url_parse(&url, url_str)) != 0) {
 		return (rv);
 	}
 	if (((tran = nni_sp_tran_find(url)) == NULL) ||
@@ -246,13 +237,12 @@ nni_dialer_create(nni_dialer **dp, nni_sock *s, const char *urlstr)
 		nni_url_free(url);
 		return (NNG_ENOMEM);
 	}
-	d->d_url     = url;
-	d->d_closed  = false;
-	d->d_closing = false;
-	d->d_data    = NULL;
-	d->d_ref     = 1;
-	d->d_sock    = s;
-	d->d_tran    = tran;
+	d->d_url    = url;
+	d->d_closed = false;
+	d->d_data   = NULL;
+	d->d_ref    = 1;
+	d->d_sock   = s;
+	d->d_tran   = tran;
 	nni_atomic_flag_reset(&d->d_started);
 
 	// Make a copy of the endpoint operations.  This allows us to
@@ -269,7 +259,7 @@ nni_dialer_create(nni_dialer **dp, nni_sock *s, const char *urlstr)
 	nni_aio_init(&d->d_tmo_aio, dialer_timer_cb, d);
 
 	nni_mtx_lock(&dialers_lk);
-	rv = nni_id_alloc(&dialers, &d->d_id, d);
+	rv = nni_id_alloc32(&dialers, &d->d_id, d);
 	nni_mtx_unlock(&dialers_lk);
 
 #ifdef NNG_ENABLE_STATS
@@ -332,6 +322,7 @@ nni_dialer_rele(nni_dialer *d)
 	bool reap;
 
 	nni_mtx_lock(&dialers_lk);
+	NNI_ASSERT(d->d_ref > 0);
 	d->d_ref--;
 	reap = ((d->d_ref == 0) && (d->d_closed));
 	nni_mtx_unlock(&dialers_lk);
@@ -341,22 +332,6 @@ nni_dialer_rele(nni_dialer *d)
 	}
 }
 
-void
-nni_dialer_close_rele(nni_dialer *d)
-{
-	nni_mtx_lock(&dialers_lk);
-	if (d->d_closed) {
-		nni_mtx_unlock(&dialers_lk);
-		nni_dialer_rele(d);
-		return;
-	}
-	d->d_closed = true;
-	nni_id_remove(&dialers, d->d_id);
-	nni_mtx_unlock(&dialers_lk);
-
-	nni_dialer_rele(d);
-}
-
 void
 nni_dialer_close(nni_dialer *d)
 {
@@ -372,6 +347,7 @@ nni_dialer_close(nni_dialer *d)
 
 	nni_dialer_shutdown(d);
 
+	nni_sock_remove_dialer(d);
 	nni_dialer_rele(d);
 }
 
@@ -389,8 +365,8 @@ static void
 dialer_connect_cb(void *arg)
 {
 	nni_dialer *d   = arg;
-	nni_aio *   aio = &d->d_con_aio;
-	nni_aio *   user_aio;
+	nni_aio    *aio = &d->d_con_aio;
+	nni_aio    *user_aio;
 	int         rv;
 
 	nni_mtx_lock(&d->d_mtx);
@@ -412,6 +388,11 @@ dialer_connect_cb(void *arg)
 	case NNG_ECONNREFUSED:
 	case NNG_ETIMEDOUT:
 	default:
+		nng_log_warn("NNG-CONN-FAIL",
+		    "Failed connecting socket<%u> to %s: %s",
+		    nni_sock_id(d->d_sock), d->d_url->u_rawurl,
+		    nng_strerror(rv));
+
 		nni_dialer_bump_error(d, rv);
 		if (user_aio == NULL) {
 			nni_dialer_timer_start(d);
@@ -434,8 +415,8 @@ dialer_connect_start(nni_dialer *d)
 int
 nni_dialer_start(nni_dialer *d, unsigned flags)
 {
-	int      rv = 0;
-	nni_aio *aio;
+	int      rv  = 0;
+	nni_aio *aio = NULL;
 
 	if (nni_atomic_flag_test_and_set(&d->d_started)) {
 		return (NNG_ESTATE);
@@ -462,9 +443,20 @@ nni_dialer_start(nni_dialer *d, unsigned flags)
 		nni_aio_free(aio);
 	}
 
+	nng_log_info("NNG-DIAL", "Starting dialer for socket<%u> on %s",
+	    nni_sock_id(d->d_sock), d->d_url->u_rawurl);
+
 	return (rv);
 }
 
+void
+nni_dialer_stop(nni_dialer *d)
+{
+	nni_aio_stop(&d->d_tmo_aio);
+	nni_aio_stop(&d->d_con_aio);
+	d->d_ops.d_close(d->d_data);
+}
+
 nni_sock *
 nni_dialer_sock(nni_dialer *d)
 {
