@@ -1,5 +1,5 @@
 //
-// Copyright 2021 Staysail Systems, Inc. <info@staysail.tech>
+// Copyright 2024 Staysail Systems, Inc. <info@staysail.tech>
 // Copyright 2018 Capitar IT Group BV <info@capitar.com>
 // Copyright 2018 Devolutions <info@devolutions.net>
 //
@@ -10,6 +10,8 @@
 //
 
 #include "core/nng_impl.h"
+#include "core/strs.h"
+#include "nng/nng.h"
 #include "sockimpl.h"
 
 #include <stdio.h>
@@ -21,25 +23,8 @@ static void listener_accept_start(nni_listener *);
 static void listener_accept_cb(void *);
 static void listener_timer_cb(void *);
 
-static nni_id_map listeners;
-static nni_mtx    listeners_lk;
-
-int
-nni_listener_sys_init(void)
-{
-	nni_id_map_init(&listeners, 1, 0x7fffffff, false);
-	nni_mtx_init(&listeners_lk);
-
-	return (0);
-}
-
-void
-nni_listener_sys_fini(void)
-{
-	nni_reap_drain();
-	nni_mtx_fini(&listeners_lk);
-	nni_id_map_fini(&listeners);
-}
+static nni_id_map listeners    = NNI_ID_MAP_INITIALIZER(1, 0x7fffffff, 0);
+static nni_mtx    listeners_lk = NNI_MTX_INITIALIZER;
 
 uint32_t
 nni_listener_id(nni_listener *l)
@@ -50,9 +35,7 @@ nni_listener_id(nni_listener *l)
 void
 nni_listener_destroy(nni_listener *l)
 {
-	nni_aio_stop(&l->l_acc_aio);
-	nni_aio_stop(&l->l_tmo_aio);
-
+	// NB: both these will have already been stopped.
 	nni_aio_fini(&l->l_acc_aio);
 	nni_aio_fini(&l->l_tmo_aio);
 
@@ -216,13 +199,18 @@ nni_listener_bump_error(nni_listener *l, int err)
 #endif
 }
 
+// nni_listener_create creates a listener on the socket.
+// The caller should have a hold on the socket, and on success
+// the listener inherits the callers hold.  (If the caller wants
+// an additional hold, it should get an extra hold before calling this
+// function.)
 int
 nni_listener_create(nni_listener **lp, nni_sock *s, const char *url_str)
 {
-	nni_sp_tran *    tran;
+	nni_sp_tran  *tran;
 	nni_listener *l;
 	int           rv;
-	nni_url *     url;
+	nni_url      *url;
 
 	if ((rv = nni_url_parse(&url, url_str)) != 0) {
 		return (rv);
@@ -237,13 +225,12 @@ nni_listener_create(nni_listener **lp, nni_sock *s, const char *url_str)
 		nni_url_free(url);
 		return (NNG_ENOMEM);
 	}
-	l->l_url     = url;
-	l->l_closed  = false;
-	l->l_closing = false;
-	l->l_data    = NULL;
-	l->l_ref     = 1;
-	l->l_sock    = s;
-	l->l_tran    = tran;
+	l->l_url    = url;
+	l->l_closed = false;
+	l->l_data   = NULL;
+	l->l_ref    = 1;
+	l->l_sock   = s;
+	l->l_tran   = tran;
 	nni_atomic_flag_reset(&l->l_started);
 
 	// Make a copy of the endpoint operations.  This allows us to
@@ -258,7 +245,7 @@ nni_listener_create(nni_listener **lp, nni_sock *s, const char *url_str)
 	nni_aio_init(&l->l_tmo_aio, listener_timer_cb, l);
 
 	nni_mtx_lock(&listeners_lk);
-	rv = nni_id_alloc(&listeners, &l->l_id, l);
+	rv = nni_id_alloc32(&listeners, &l->l_id, l);
 	nni_mtx_unlock(&listeners_lk);
 
 #ifdef NNG_ENABLE_STATS
@@ -321,6 +308,7 @@ nni_listener_rele(nni_listener *l)
 	bool reap;
 
 	nni_mtx_lock(&listeners_lk);
+	NNI_ASSERT(l->l_ref > 0);
 	l->l_ref--;
 	reap = ((l->l_ref == 0) && (l->l_closed));
 	nni_mtx_unlock(&listeners_lk);
@@ -344,24 +332,8 @@ nni_listener_close(nni_listener *l)
 
 	nni_listener_shutdown(l);
 
-	nni_listener_rele(l); // This will trigger a reap if id count is zero.
-}
-
-void
-nni_listener_close_rele(nni_listener *l)
-{
-	// Listener should already be shutdown.  The socket lock may be held.
-	nni_mtx_lock(&listeners_lk);
-	if (l->l_closed) {
-		nni_mtx_unlock(&listeners_lk);
-		nni_listener_rele(l);
-		return;
-	}
-	l->l_closed = true;
-	nni_id_remove(&listeners, l->l_id);
-	nni_mtx_unlock(&listeners_lk);
-
-	nni_listener_rele(l); // This will trigger a reap if id count is zero.
+	nni_sock_remove_listener(l);
+	nni_listener_rele(l); // This will reap if reference count is zero.
 }
 
 static void
@@ -378,7 +350,7 @@ static void
 listener_accept_cb(void *arg)
 {
 	nni_listener *l   = arg;
-	nni_aio *     aio = &l->l_acc_aio;
+	nni_aio      *aio = &l->l_acc_aio;
 	int           rv;
 
 	switch ((rv = nni_aio_result(aio))) {
@@ -393,6 +365,10 @@ listener_accept_cb(void *arg)
 	case NNG_ECONNRESET:   // remote condition, no cool down
 	case NNG_ETIMEDOUT:    // No need to sleep, we timed out already.
 	case NNG_EPEERAUTH:    // peer validation failure
+		nng_log_warn("NNG-ACCEPT-FAIL",
+		    "Failed accepting for socket<%u> on %s: %s",
+		    nni_sock_id(l->l_sock), l->l_url->u_rawurl,
+		    nng_strerror(rv));
 		nni_listener_bump_error(l, rv);
 		listener_accept_start(l);
 		break;
@@ -422,7 +398,9 @@ listener_accept_start(nni_listener *l)
 int
 nni_listener_start(nni_listener *l, int flags)
 {
-	int rv;
+	int    rv;
+	char  *url;
+	size_t sz;
 	NNI_ARG_UNUSED(flags);
 
 	if (nni_atomic_flag_test_and_set(&l->l_started)) {
@@ -430,16 +408,35 @@ nni_listener_start(nni_listener *l, int flags)
 	}
 
 	if ((rv = l->l_ops.l_bind(l->l_data)) != 0) {
+		nng_log_warn("NNG-BIND-FAIL",
+		    "Failed binding socket<%u> to %s: %s",
+		    nni_sock_id(l->l_sock), l->l_url->u_rawurl,
+		    nng_strerror(rv));
 		nni_listener_bump_error(l, rv);
 		nni_atomic_flag_reset(&l->l_started);
 		return (rv);
 	}
+	// collect the URL which may have changed (e.g. binding to port 0)
+	sz = sizeof(url);
+	(void) (nni_listener_getopt(
+	    l, NNG_OPT_URL, &url, &sz, NNI_TYPE_STRING));
+	nng_log_info("NNG-LISTEN", "Starting listener for socket<%u> on %s",
+	    nni_sock_id(l->l_sock), url);
+	nni_strfree(url);
 
 	listener_accept_start(l);
 
 	return (0);
 }
 
+void
+nni_listener_stop(nni_listener *l)
+{
+	nni_aio_stop(&l->l_tmo_aio);
+	nni_aio_stop(&l->l_acc_aio);
+	l->l_ops.l_close(l->l_data);
+}
+
 nni_sock *
 nni_listener_sock(nni_listener *l)
 {
@@ -513,7 +510,7 @@ nni_listener_getopt(
 void
 nni_listener_add_stat(nni_listener *l, nni_stat_item *item)
 {
-#if NNG_ENABLE_STATS
+#ifdef NNG_ENABLE_STATS
 	nni_stat_add(&l->st_root, item);
 #else
 	NNI_ARG_UNUSED(l);
